import re

from wfuzz.plugin_api.base import BasePlugin
from wfuzz.externals.moduleman.plugin import moduleman_plugin


@moduleman_plugin
class errors(BasePlugin):
    name = "errors"
    author = ("Xavi Mendez (@xmendez)",)
    version = "0.1"
    summary = "Looks for error messages"
    description = ("Looks for common error messages",)
    category = ["default", "passive", "info"]
    priority = 99

    parameters = ()

    def __init__(self):
        BasePlugin.__init__(self)

        regex_list = [
            "A syntax error has occurred",
            "ADODB.Field error",
            "ASP.NET is configured to show verbose error messages",
            "ASP.NET_SessionId",
            "Active Server Pages error",
            "An illegal character has been found in the statement",
            'An unexpected token "END-OF-STATEMENT" was found',
            "Can't connect to local",
            "Custom Error Message",
            "DB2 Driver",
            "DB2 Error",
            "DB2 ODBC",
            "Disallowed Parent Path",
            "Error Diagnostic Information",
            "Error Message : Error loading required libraries.",
            "Error Report",
            "Error converting data type varchar to numeric",
            "Fatal error",
            "Incorrect syntax near",
            "Internal Server Error",
            "Invalid Path Character",
            "Invalid procedure call or argument",
            "Invision Power Board Database Error",
            "JDBC Driver",
            "JDBC Error",
            "JDBC MySQL",
            "JDBC Oracle",
            "JDBC SQL",
            "Microsoft OLE DB Provider for ODBC Drivers",
            "Microsoft VBScript compilation error",
            "Microsoft VBScript error",
            "MySQL Driver",
            "MySQL Error",
            "MySQL ODBC",
            "ODBC DB2",
            "ODBC Driver",
            "ODBC Error",
            "ODBC Microsoft Access",
            "ODBC Oracle",
            "ODBC SQL",
            "ODBC SQL Server",
            "OLE/DB provider returned message",
            "ORA-0",
            "ORA-1",
            "Oracle DB2",
            "Oracle Driver",
            "Oracle Error",
            "Oracle ODBC",
            "PHP Error",
            "PHP Parse error",
            "PHP Warning",
            "Permission denied: 'GetObject'",
            "PostgreSQL query failed: ERROR: parser: parse error",
            r"SQL Server Driver\]\[SQL Server",
            "SQL command not properly ended",
            "SQLException",
            "Supplied argument is not a valid PostgreSQL result",
            "Syntax error in query expression",
            "The error occurred in",
            "The script whose uid is",
            "Type mismatch",
            "Unable to jump to row",
            "Unclosed quotation mark before the character string",
            "Unterminated string constant",
            "Warning: Cannot modify header information - headers already sent",
            "Warning: Supplied argument is not a valid File-Handle resource in",
            r"Warning: mysql_query\(\)",
            r"Warning: mysql_fetch_array\(\)",
            r"Warning: pg_connect\(\): Unable to connect to PostgreSQL server: FATAL",
            "You have an error in your SQL syntax near",
            "data source=",
            "detected an internal error [IBM][CLI Driver][DB2/6000]",
            "invalid query",
            "is not allowed to access",
            "missing expression",
            "mySQL error with query",
            "mysql error",
            "on MySQL result index",
            "supplied argument is not a valid MySQL result resource",
        ]

        self.error_regex = []
        for regex in regex_list:
            self.error_regex.append(re.compile(regex, re.MULTILINE | re.DOTALL))

    def validate(self, fuzzresult):
        return True

    def process(self, fuzzresult):
        for regex in self.error_regex:
            for regex_match in regex.findall(fuzzresult.history.content):
                self.add_result("errors", "Error identified", regex_match)



import re
import cgi

from io import BytesIO
import gzip
import zlib

from .TextParser import TextParser

from wfuzz.helpers.str_func import python2_3_convert_from_unicode


def get_encoding_from_headers(headers):
    """Returns encodings from given HTTP Header Dict.

    :param headers: dictionary to extract encoding from.
    :rtype: str
    """

    content_type = headers.get("Content-Type")

    if not content_type:
        return None

    content_type, params = cgi.parse_header(content_type)

    if "charset" in params:
        return params["charset"].strip("'\"")

    if "text" in content_type:
        return "ISO-8859-1"

    if "image" in content_type:
        return "utf-8"

    if "application/json" in content_type:
        return "utf-8"


def get_encodings_from_content(content):
    """Returns encodings from given content string.

    :param content: bytestring to extract encodings from.
    """
    charset_re = re.compile(r'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
    pragma_re = re.compile(r'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
    xml_re = re.compile(r'^<\?xml.*?encoding=["\']*(.+?)["\'>]')

    return (
        charset_re.findall(content)
        + pragma_re.findall(content)
        + xml_re.findall(content)
    )


class Response:
    def __init__(self, protocol="", code="", message=""):
        self.protocol = protocol  # HTTP/1.1
        self.code = code  # 200
        self.message = message  # OK
        self._headers = []  # bueno pues las cabeceras igual que en la request
        self.__content = (
            ""  # contenido de la response (si i solo si Content-Length existe)
        )
        self.md5 = ""  # hash de los contenidos del resultado
        self.charlen = ""  # Cantidad de caracteres de la respuesta

    def addHeader(self, key, value):
        self._headers += [(key, value)]

    def delHeader(self, key):
        for i in self._headers:
            if i[0].lower() == key.lower():
                self._headers.remove(i)

    def addContent(self, text):
        self.__content = self.__content + text

    def __getitem__(self, key):
        for i, j in self._headers:
            if key == i:
                return j
        print("Error al obtener header!!!")

    def getCookie(self):
        str = []
        for i, j in self._headers:
            if i.lower() == "set-cookie":
                str.append(j.split(";")[0])
        return "; ".join(str)

    def has_header(self, key):
        for i, j in self._headers:
            if i.lower() == key.lower():
                return True
        return False

    def getLocation(self):
        for i, j in self._headers:
            if i.lower() == "location":
                return j
        return None

    def header_equal(self, header, value):
        for i, j in self._headers:
            if i == header and j.lower() == value.lower():
                return True
        return False

    def getHeaders(self):
        return self._headers

    def getContent(self):
        return self.__content

    def getTextHeaders(self):
        string = (
            str(self.protocol) + " " + str(self.code) + " " + str(self.message) + "\r\n"
        )
        for i, j in self._headers:
            string += i + ": " + j + "\r\n"

        return string

    def getAll(self):
        string = self.getTextHeaders() + "\r\n" + self.getContent()
        return string

    def Substitute(self, src, dst):
        a = self.getAll()
        b = a.replace(src, dst)
        self.parseResponse(b)

    def getAll_wpost(self):
        string = (
            str(self.protocol) + " " + str(self.code) + " " + str(self.message) + "\r\n"
        )
        for i, j in self._headers:
            string += i + ": " + j + "\r\n"
        return string

    def parseResponse(self, rawheader, rawbody=None, type="curl"):
        self.__content = ""
        self._headers = []

        tp = TextParser()
        tp.setSource("string", rawheader)

        tp.readUntil(r"(HTTP/[0-9.]+) ([0-9]+)")
        while True:
            while True:
                try:
                    self.protocol = tp[0][0]
                except Exception:
                    self.protocol = "unknown"

                try:
                    self.code = tp[0][1]
                except Exception:
                    self.code = "0"

                if self.code != "100":
                    break
                else:
                    tp.readUntil(r"(HTTP/[0-9.]+) ([0-9]+)")

            self.code = int(self.code)

            while True:
                tp.readLine()
                if tp.search("^([^:]+): ?(.*)$"):
                    self.addHeader(tp[0][0], tp[0][1])
                else:
                    break

            # curl sometimes sends two headers when using follow, 302 and the final header
            # also when using proxies
            tp.readLine()
            if not tp.search(r"(HTTP/[0-9.]+) ([0-9]+)"):
                break
            else:
                self._headers = []

        # ignore CRLFs until request line
        while tp.lastline == "" and tp.readLine():
            pass

        # TODO: this should be added to rawbody not directly to __content
        if tp.lastFull_line:
            self.addContent(tp.lastFull_line)

        while tp.skip(1):
            self.addContent(tp.lastFull_line)

        if type == "curl":
            self.delHeader("Transfer-Encoding")

        if self.header_equal("Transfer-Encoding", "chunked"):
            result = ""
            content = BytesIO(rawbody)
            hexa = content.readline()
            nchunk = int(hexa.strip(), 16)

            while nchunk:
                result += content.read(nchunk)
                content.readline()
                hexa = content.readline()
                nchunk = int(hexa.strip(), 16)

            rawbody = result

        if self.header_equal("Content-Encoding", "gzip"):
            compressedstream = BytesIO(rawbody)
            gzipper = gzip.GzipFile(fileobj=compressedstream)
            rawbody = gzipper.read()
            self.delHeader("Content-Encoding")
        elif self.header_equal("Content-Encoding", "deflate"):
            deflated_data = None
            try:
                deflater = zlib.decompressobj()
                deflated_data = deflater.decompress(rawbody)
                deflated_data += deflater.flush()
            except zlib.error:
                try:
                    deflater = zlib.decompressobj(-zlib.MAX_WBITS)
                    deflated_data = deflater.decompress(rawbody)
                    deflated_data += deflater.flush()
                except zlib.error:
                    deflated_data = ""
            rawbody = deflated_data
            self.delHeader("Content-Encoding")

        if rawbody is not None:
            # Try to get charset encoding from headers
            content_encoding = get_encoding_from_headers(dict(self.getHeaders()))

            # fallback to default encoding
            if content_encoding is None:
                content_encoding = "utf-8"

            self.__content = python2_3_convert_from_unicode(
                rawbody.decode(content_encoding, errors="replace")
            )



import re
import sys
import six


from .obj_dic import DotDict


def json_minify(string, strip_space=True):
    """
    Created on 20/01/2011
    v0.2 (C) Gerald Storer
    MIT License
    Based on JSON.minify.js:
    https://github.com/getify/JSON.minify
    Contributers:
    - Pradyun S. Gedam (conditions and variable names changed)
    """

    tokenizer = re.compile(r'"|(/\*)|(\*/)|(//)|\n|\r')
    end_slashes_re = re.compile(r"(\\)*$")

    in_string = False
    in_multi = False
    in_single = False

    new_str = []
    index = 0

    for match in re.finditer(tokenizer, string):

        if not (in_multi or in_single):
            tmp = string[index : match.start()]
            if not in_string and strip_space:
                # replace white space as defined in standard
                tmp = re.sub("[ \t\n\r]+", "", tmp)
            new_str.append(tmp)

        index = match.end()
        val = match.group()

        if val == '"' and not (in_multi or in_single):
            escaped = end_slashes_re.search(string, 0, match.start())

            # start of string or unescaped quote character to end string
            if not in_string or (escaped is None or len(escaped.group()) % 2 == 0):
                in_string = not in_string
            # include " character in next catch
            index -= 1
        elif not (in_string or in_multi or in_single):
            if val == "/*":
                in_multi = True
            elif val == "//":
                in_single = True
        elif val == "*/" and in_multi and not (in_string or in_single):
            in_multi = False
        elif val in "\r\n" and not (in_multi or in_string) and in_single:
            in_single = False
        elif not ((in_multi or in_single) or (val in " \r\n\t" and strip_space)):
            new_str.append(val)

    new_str.append(string[index:])
    return "".join(new_str)


def python2_3_convert_from_unicode(text):
    if sys.version_info >= (3, 0):
        return text
    else:
        return convert_to_unicode(text)


def python2_3_convert_to_unicode(text):
    if sys.version_info >= (3, 0):
        return convert_to_unicode(text)
    else:
        return text


def convert_to_unicode(text):
    if isinstance(text, dict) or isinstance(text, DotDict):
        return {
            convert_to_unicode(key): convert_to_unicode(value)
            for key, value in list(text.items())
        }
    elif isinstance(text, list):
        return [convert_to_unicode(element) for element in text]
    elif isinstance(text, six.string_types):
        return text.encode("utf-8", errors="ignore")
    else:
        return text


def value_in_any_list_item(value, list_obj):
    if isinstance(list_obj, list):
        return len([item for item in list_obj if value.lower() in item.lower()]) > 0
    elif isinstance(list_obj, str):
        return value.lower() in list_obj.lower()



import pycurl
from io import BytesIO
from threading import Thread, Lock
import itertools
from queue import Queue
import collections

from .exception import FuzzExceptBadOptions, FuzzExceptNetError

from .factories.reqresp_factory import ReqRespRequestFactory

# See https://curl.haxx.se/libcurl/c/libcurl-errors.html
UNRECOVERABLE_PYCURL_EXCEPTIONS = [
    28,  # Operation timeout. The specified time-out period was reached according to the conditions.
    7,  # Failed to connect() to host or proxy.
    6,  # Couldn't resolve host. The given remote host was not resolved.
    5,  # Couldn't resolve proxy. The given proxy host could not be resolved.
]

# Other common pycurl exceptions:
# Exception in perform (35, 'error:0B07C065:x509 certificate routines:X509_STORE_add_cert:cert already in hash table')
# Exception in perform (18, 'SSL read: error:0B07C065:x509 certificate routines:X509_STORE_add_cert:cert already in hash table, errno 11')


class HttpPool:
    HTTPAUTH_BASIC, HTTPAUTH_NTLM, HTTPAUTH_DIGEST = ("basic", "ntlm", "digest")
    newid = itertools.count(0)

    def __init__(self, options):
        self.processed = 0

        self.exit_job = False
        self.mutex_stats = Lock()

        self.m = None
        self.curlh_freelist = []
        self._request_list = collections.deque()
        self.handles = []

        self.ths = None

        self.pool_map = {}

        self.options = options

        self._registered = 0

    def _initialize(self):
        # pycurl Connection pool
        self.m = pycurl.CurlMulti()
        self.handles = []

        for i in range(self.options.get("concurrent")):
            curl_h = pycurl.Curl()
            self.handles.append(curl_h)
            self.curlh_freelist.append(curl_h)

        # create threads
        self.ths = []

        for fn in ("_read_multi_stack",):
            th = Thread(target=getattr(self, fn))
            th.setName(fn)
            self.ths.append(th)
            th.start()

    def job_stats(self):
        with self.mutex_stats:
            dic = {
                "http_processed": self.processed,
                "http_registered": self._registered,
            }
        return dic

    # internal http pool control

    def iter_results(self, poolid):
        item = self.pool_map[poolid]["queue"].get()

        if not item:
            return

        yield item

    def _new_pool(self):
        poolid = next(self.newid)
        self.pool_map[poolid] = {}
        self.pool_map[poolid]["queue"] = Queue()
        self.pool_map[poolid]["proxy"] = None

        if self.options.get("proxies"):
            self.pool_map[poolid]["proxy"] = self._get_next_proxy(
                self.options.get("proxies")
            )

        return poolid

    def _prepare_curl_h(self, curl_h, fuzzres, poolid):
        new_curl_h = ReqRespRequestFactory.to_http_object(
            self.options, fuzzres.history, curl_h
        )
        new_curl_h = self._set_extra_options(new_curl_h, fuzzres, poolid)

        new_curl_h.response_queue = (BytesIO(), BytesIO(), fuzzres, poolid)
        new_curl_h.setopt(pycurl.WRITEFUNCTION, new_curl_h.response_queue[0].write)
        new_curl_h.setopt(pycurl.HEADERFUNCTION, new_curl_h.response_queue[1].write)

        return new_curl_h

    def enqueue(self, fuzzres, poolid):
        if self.exit_job:
            return

        self._request_list.append((fuzzres, poolid))

    def _stop_to_pools(self):
        for p in list(self.pool_map.keys()):
            self.pool_map[p]["queue"].put(None)

    def cleanup(self):
        self.exit_job = True
        for th in self.ths:
            th.join()

    def register(self):
        with self.mutex_stats:
            self._registered += 1

        if not self.pool_map:
            self._initialize()

        return self._new_pool()

    def deregister(self):
        with self.mutex_stats:
            self._registered -= 1

            if self._registered <= 0:
                self.cleanup()

    def _get_next_proxy(self, proxy_list):
        i = 0
        while 1:
            yield proxy_list[i]
            i += 1
            i = i % len(proxy_list)

    def _set_extra_options(self, c, fuzzres, poolid):
        if self.pool_map[poolid]["proxy"]:
            ip, port, ptype = next(self.pool_map[poolid]["proxy"])

            fuzzres.history.wf_proxy = (("%s:%s" % (ip, port)), ptype)

            if ptype == "SOCKS5":
                c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)
                c.setopt(pycurl.PROXY, "%s:%s" % (ip, port))
            elif ptype == "SOCKS4":
                c.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS4)
                c.setopt(pycurl.PROXY, "%s:%s" % (ip, port))
            elif ptype == "HTTP":
                c.setopt(pycurl.PROXY, "%s:%s" % (ip, port))
            else:
                raise FuzzExceptBadOptions(
                    "Bad proxy type specified, correct values are HTTP, SOCKS4 or SOCKS5."
                )
        else:
            c.setopt(pycurl.PROXY, "")

        mdelay = self.options.get("req_delay")
        if mdelay is not None:
            c.setopt(pycurl.TIMEOUT, mdelay)

        cdelay = self.options.get("conn_delay")
        if cdelay is not None:
            c.setopt(pycurl.CONNECTTIMEOUT, cdelay)

        return c

    def _process_curl_handle(self, curl_h):
        buff_body, buff_header, res, poolid = curl_h.response_queue

        try:
            ReqRespRequestFactory.from_http_object(
                self.options,
                res.history,
                curl_h,
                buff_header.getvalue(),
                buff_body.getvalue(),
            )
        except Exception as e:
            self.pool_map[poolid]["queue"].put(res.update(exception=e))
        else:
            # reset type to result otherwise backfeed items will enter an infinite loop
            self.pool_map[poolid]["queue"].put(res.update())

        with self.mutex_stats:
            self.processed += 1

    def _process_curl_should_retry(self, res, errno, poolid):
        if errno not in UNRECOVERABLE_PYCURL_EXCEPTIONS:
            res.history.wf_retries += 1

            if res.history.wf_retries < self.options.get("retries"):
                self._request_list.append((res, poolid))
                return True

        return False

    def _process_curl_handle_error(self, res, errno, errmsg, poolid):
        e = FuzzExceptNetError("Pycurl error %d: %s" % (errno, errmsg))
        res.history.totaltime = 0
        self.pool_map[poolid]["queue"].put(res.update(exception=e))

        with self.mutex_stats:
            self.processed += 1

    def _read_multi_stack(self):
        # Check for curl objects which have terminated, and add them to the curlh_freelist
        while not self.exit_job:
            while not self.exit_job:
                ret, num_handles = self.m.perform()
                if ret != pycurl.E_CALL_MULTI_PERFORM:
                    break

            num_q, ok_list, err_list = self.m.info_read()
            for curl_h in ok_list:
                self._process_curl_handle(curl_h)
                self.m.remove_handle(curl_h)
                self.curlh_freelist.append(curl_h)

            for curl_h, errno, errmsg in err_list:
                buff_body, buff_header, res, poolid = curl_h.response_queue

                if not self._process_curl_should_retry(res, errno, poolid):
                    self._process_curl_handle_error(res, errno, errmsg, poolid)

                self.m.remove_handle(curl_h)
                self.curlh_freelist.append(curl_h)

            while self.curlh_freelist and self._request_list:
                curl_h = self.curlh_freelist.pop()
                fuzzres, poolid = self._request_list.popleft()

                self.m.add_handle(self._prepare_curl_h(curl_h, fuzzres, poolid))

        self._stop_to_pools()

        # cleanup multi stack
        for c in self.handles:
            c.close()
            self.curlh_freelist.append(c)
        self.m.close()



from .fuzzobjects import FuzzType

from .myqueues import MyPriorityQueue, QueueManager
from .fuzzqueues import (
    SeedQ,
    SaveQ,
    PrinterQ,
    RoutingQ,
    FilterQ,
    SliceQ,
    JobQ,
    RecursiveQ,
    DryRunQ,
    HttpQueue,
    HttpReceiver,
    AllVarQ,
    CLIPrinterQ,
    ConsolePrinterQ,
    PassPayloadQ,
)


# python 2 and 3: iterator
from builtins import object


class Fuzzer(object):
    def __init__(self, options):
        # Create queues
        # genReq ---> seed_queue -> [slice_queue] -> http_queue/dryrun -> [round_robin -> plugins_queue] * N
        # -> [recursive_queue -> routing_queue] -> [filter_queue] -> [save_queue] -> [printer_queue] ---> results

        self.options = options
        self.qmanager = QueueManager(options)
        self.results_queue = MyPriorityQueue()

        if options["allvars"]:
            self.qmanager.add("seed_queue", AllVarQ(options))
        else:
            self.qmanager.add("seed_queue", SeedQ(options))

        for prefilter_idx, prefilter in enumerate(options.get("compiled_prefilter")):
            if prefilter.is_active():
                self.qmanager.add(
                    "slice_queue_{}".format(prefilter_idx), SliceQ(options, prefilter)
                )

        if options.get("transport") == "dryrun":
            self.qmanager.add("transport_queue", DryRunQ(options))
        elif options.get("transport") == "payload":
            self.qmanager.add("transport_queue", PassPayloadQ(options))
        else:
            # http_queue breaks process rules due to being asynchronous. Someone has to collects its sends, for proper fuzzqueue's count and sync purposes
            self.qmanager.add("transport_queue", HttpQueue(options))
            self.qmanager.add("http_receiver", HttpReceiver(options))

        if options.get("script"):
            self.qmanager.add("plugins_queue", JobQ(options))

        if options.get("rlevel") > 0:
            self.qmanager.add("recursive_queue", RecursiveQ(options))

        if (options.get("script") or options.get("rlevel") > 0) and options.get(
            "transport"
        ) == "http":
            rq = RoutingQ(
                options,
                {
                    FuzzType.SEED: self.qmanager["seed_queue"],
                    FuzzType.BACKFEED: self.qmanager["transport_queue"],
                },
            )

            self.qmanager.add("routing_queue", rq)

        if options.get("compiled_filter").is_active():
            self.qmanager.add(
                "filter_queue", FilterQ(options, options["compiled_filter"])
            )

        if options.get("compiled_simple_filter").is_active():
            self.qmanager.add(
                "simple_filter_queue",
                FilterQ(options, options["compiled_simple_filter"]),
            )

        if options.get("save"):
            self.qmanager.add("save_queue", SaveQ(options))

        if options.get("compiled_printer"):
            self.qmanager.add("printer_queue", PrinterQ(options))

        if options.get("exec_mode") == "cli":
            if options["console_printer"]:
                self.qmanager.add("printer_cli", ConsolePrinterQ(options))
            else:
                self.qmanager.add("printer_cli", CLIPrinterQ(options))

        self.qmanager.bind(self.results_queue)

        # initial seed request
        self.qmanager.start()

    def __iter__(self):
        return self

    def __next__(self):
        # http://bugs.python.org/issue1360
        res = self.results_queue.get()
        self.results_queue.task_done()

        # done! (None sent has gone through all queues).
        if not res:
            raise StopIteration
        elif res.item_type == FuzzType.ERROR:
            raise res.exception

        return res

    def stats(self):
        return dict(
            list(self.qmanager.get_stats().items())
            + list(self.qmanager["transport_queue"].http_pool.job_stats().items())
            + list(self.options.stats.get_stats().items())
        )

    def cancel_job(self):
        self.qmanager.cancel()

    def pause_job(self):
        self.qmanager["transport_queue"].pause.clear()

    def resume_job(self):
        self.qmanager["transport_queue"].pause.set()



from .modulefilter import Filter
from collections import defaultdict

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
from threading import Lock


class IRegistrant:
    def __init__(self, loader, plg_filter):
        self.plg_filter = plg_filter
        self.loader = loader

        self.start_loading()
        self.load()
        self.end_loading()

    def register(self, identifier, module):
        raise NotImplementedError

    def start_loading(self):
        raise NotImplementedError

    def load(self):
        raise NotImplementedError

    def end_loading(self):
        raise NotImplementedError

    def modify_instance(self, module):
        raise NotImplementedError


class KnowledgeBase(MutableMapping):
    def __init__(self, *args, **kwargs):
        self.__data = defaultdict(list)
        self.mutex = Lock()

    def __getitem__(self, key):
        with self.mutex:
            return self.__data[key]

    def __setitem__(self, key, value):
        with self.mutex:
            self.__data[key].append(value)

    def __delitem__(self, key):
        with self.mutex:
            del self.__data[key]

    def __len__(self):
        with self.mutex:
            return len(self.__data)

    def __str__(self):
        with self.mutex:
            return str(self.__data)

    def __iter__(self):
        return iter(self.__data)


class BRegistrant(IRegistrant):
    def __init__(self, loader, plg_filter=Filter()):
        self.__plugins = {}
        self.__active_plugins = {}
        self.kbase = KnowledgeBase()

        IRegistrant.__init__(self, loader, plg_filter)

    def register(self, identifier, module):
        self.__plugins[identifier] = self.modify_instance(module)
        self.__active_plugins[identifier] = True

    def load(self):
        self.loader.load(self)

    def start_loading(self):
        pass

    def end_loading(self):
        pass

    def modify_instance(self, module):
        module.kbase = self.kbase

        return module

    # ------------------------------------------------
    # plugin management functions
    # ------------------------------------------------
    def plugin_state(self, identifier, state):
        self.__active_plugins[identifier] = state

    def __get_plugins(self, category, sorting):
        def plugin_filter(x):
            plgid, plg = x

            if category == "$all$":
                return True
            elif not self.__active_plugins[plgid]:
                return False
            else:
                return self.plg_filter.is_visible(plg, category)

        def key_funtion(x):
            return x[1].priority

        plugin_list = list(filter(plugin_filter, list(self.__plugins.items())))

        if sorting:
            plugin_list.sort(key=key_funtion)

        return plugin_list

    def get_plugin(self, identifier):
        # strict and fuzzy search
        if identifier in self.__plugins:
            return self.__plugins[identifier]
        else:
            plugin_list = [
                plg
                for plg_id, plg in self.__get_plugins("$all$", True)
                if identifier in plg_id
            ]

            if not plugin_list:
                raise KeyError("No plugins found!")
            elif len(plugin_list) == 1:
                return plugin_list[0]
            else:
                raise KeyError(
                    "Multiple plugins found: %s"
                    % ",".join([plg.name for plg in plugin_list])
                )

        raise KeyError("No plugins found!")

    def get_plugins(self, category="$all$", sorting="true"):
        return [plg for plg_id, plg in self.__get_plugins(category, sorting)]

    def get_plugins_ext(self, category="$all$", sorting="true"):
        plugin_list = [["Id", "Priority", "Category", "Name", "Summary"]]

        for plg_id, plg in self.__get_plugins(category, sorting):
            plugin_list.append(
                [
                    plg_id,
                    str(plg.priority),
                    ", ".join(plg.category),
                    str(plg.name),
                    str(plg.summary),
                ]
            )

        return plugin_list

    def get_plugins_names(self, category="$all$", sorting="true"):
        return [plg.name for plg_id, plg in self.__get_plugins(category, sorting)]

    def get_plugins_ids(self, category="$all$", sorting="true"):
        return [plg_id for plg_id, plg in self.__get_plugins(category, sorting)]


class MulRegistrant(BRegistrant):
    def load(self):
        for loader in self.loader:
            loader.load(self)



# mimicking nmap script filter


# nmap --script "http-*"
#     Loads all scripts whose name starts with http-, such as http-auth and http-open-proxy. The argument to --script had to be in quotes to protect the wildcard from the shell.
#  not valid for categories!
#
# More complicated script selection can be done using the and, or, and not operators to build Boolean expressions. The operators have the same precedence[12] as in Lua: not is the
# highest, followed by and and then or. You can alter precedence by using parentheses. Because expressions contain space characters it is necessary to quote them.
#
# nmap --script "not intrusive"
#     Loads every script except for those in the intrusive category.
#
# nmap --script "default or safe"
#     This is functionally equivalent to nmap --script "default,safe". It loads all scripts that are in the default category or the safe category or both.
#
# nmap --script "default and safe"
#     Loads those scripts that are in both the default and safe categories.
#
# nmap --script "(default or safe or intrusive) and not http-*"
#     Loads scripts in the default, safe, or intrusive categories, except for those whose names start with http-.


PYPARSING = True
try:
    from pyparsing import (
        Word,
        Group,
        oneOf,
        Optional,
        Suppress,
        ZeroOrMore,
        Literal,
        alphas,
        alphanums,
    )
except ImportError:
    PYPARSING = False


class IFilter:
    def is_visible(self, plugin, filter_string):
        raise NotImplementedError


class Filter(IFilter):
    def __init__(self):
        if PYPARSING:
            category = Word(alphas + "_-*", alphanums + "_-*")
            operator = oneOf("and or ,")
            neg_operator = "not"
            elementRef = category
            definition = elementRef + ZeroOrMore(operator + elementRef)
            nestedformula = Group(
                Suppress(Optional(Literal("(")))
                + definition
                + Suppress(Optional(Literal(")")))
            )
            neg_nestedformula = Optional(neg_operator) + nestedformula
            self.finalformula = neg_nestedformula + ZeroOrMore(
                operator + neg_nestedformula
            )

            elementRef.setParseAction(self.__compute_element)
            neg_nestedformula.setParseAction(self.__compute_neg_formula)
            nestedformula.setParseAction(self.__compute_formula)
            self.finalformula.setParseAction(self.__myreduce)

    def __compute_neg_formula(self, tokens):
        if len(tokens) > 1 and tokens[0] == "not":
            return not tokens[1]
        else:
            return tokens[0]

    def __compute_element(self, tokens):
        item = tokens[0]
        wildc_index = item.find("*")

        if wildc_index > 0:
            return self.plugin.name.startswith(item[:wildc_index])
        else:
            if isinstance(self.plugin.category, list):
                return item in self.plugin.category or self.plugin.name == item
            else:
                return self.plugin.category == item or self.plugin.name == item

    def __myreduce(self, elements):
        first = elements[0]
        for i in range(1, len(elements), 2):
            if elements[i] == "and":
                first = first and elements[i + 1]
            elif elements[i] == "or" or elements[i] == ",":
                first = first or elements[i + 1]

        return first

    def __compute_formula(self, tokens):
        return self.__myreduce(tokens[0])

    def simple_filter(self, plugin, filter_string):
        ret = []

        for item in filter_string.split(","):
            wildc_index = item.find("*")
            if wildc_index > 0:
                ret.append(
                    (
                        item in plugin.category
                        or plugin.name.startswith(item[:wildc_index])
                    )
                )
            else:
                ret.append((item in plugin.category or plugin.name == item))

        return any(ret)

    def simple_filter_banned_keywords(self, filter_string):
        if filter_string.find("(") >= 0:
            return True
        elif filter_string.find(")") >= 0:
            return True
        elif any(x in ["or", "not", "and"] for x in filter_string.split(" ")):
            return True
        else:
            return False

    def is_visible(self, plugin, filter_string):
        self.plugin = plugin
        if PYPARSING:
            return self.finalformula.parseString(filter_string)[0]
        else:
            if self.simple_filter_banned_keywords(filter_string):
                raise Exception("Pyparsing missing, complex filters not allowed.")
            else:
                return self.simple_filter(plugin, filter_string)



# python 2 and 3
from __future__ import print_function

import math
import string
import operator
from functools import reduce

# Python 2 and 3: zip_longest
from six import StringIO

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


def indent(
    rows,
    hasHeader=False,
    headerChar="-",
    delim=" | ",
    justify="left",
    separateRows=False,
    prefix="",
    postfix="",
    wrapfunc=lambda x: x,
):
    """
    @author http://code.activestate.com/recipes/267662-table-indentation/

    Indents a table by column.
    - rows: A sequence of sequences of items, one sequence per row.
    - hasHeader: True if the first row consists of the columns' names.
    - headerChar: Character to be used for the row separator line
        (if hasHeader==True or separateRows==True).
    - delim: The column delimiter.
    - justify: Determines how are data justified in their column.
        Valid values are 'left','right' and 'center'.
    - separateRows: True if rows are to be separated by a line
        of 'headerChar's.
    - prefix: A string prepended to each printed row.
    - postfix: A string appended to each printed row.
    - wrapfunc: A function f(text) for wrapping text; each element in
        the table is first wrapped by this function."""
    # closure for breaking logical rows to physical, using wrapfunc
    def rowWrapper(row):
        newRows = [wrapfunc(item).split("\n") for item in row]
        return [[substr or "" for substr in item] for item in zip_longest(*newRows)]

    # break each logical row into one or more physical ones
    logicalRows = [rowWrapper(row) for row in rows]
    # columns of physical rows
    columns = zip_longest(*reduce(operator.add, logicalRows))
    # get the maximum of each column by the string length of its items
    maxWidths = [max([len(str(item)) for item in column]) for column in columns]
    rowSeparator = headerChar * (
        len(prefix) + len(postfix) + sum(maxWidths) + len(delim) * (len(maxWidths) - 1)
    )
    # select the appropriate justify method
    justify = {"center": str.center, "right": str.rjust, "left": str.ljust}[
        justify.lower()
    ]
    output = StringIO()
    if separateRows:
        print(rowSeparator, file=output)
    for physicalRows in logicalRows:
        for row in physicalRows:
            print(
                prefix
                + delim.join(
                    [justify(str(item), width) for (item, width) in zip(row, maxWidths)]
                )
                + postfix,
                file=output,
            )
        if separateRows or hasHeader:
            print(rowSeparator, file=output)
            hasHeader = False
    return output.getvalue()


def wrap_always(text, width):
    """A simple word-wrap function that wraps text on exactly width characters.
    It doesn't split the text in words."""
    return "\n".join(
        [
            text[width * i : width * (i + 1)]
            for i in range(int(math.ceil(1.0 * len(text) / width)))
        ]
    )


def wrap_always_list(alltext, width):
    text_list = []
    for text in alltext.splitlines():
        for subtext in [
            text[width * i : width * (i + 1)]
            for i in range(int(math.ceil(1.0 * len(text) / width)))
        ]:
            text_list.append(
                "".join([char if char in string.printable else "." for char in subtext])
            )
    return text_list


def table_print(rows, width=80):
    print(
        indent(
            rows,
            hasHeader=True,
            separateRows=False,
            prefix="  ",
            postfix="  ",
            wrapfunc=lambda x: wrap_always(x, width),
        )
    )


def getTerminalSize():
    # http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
    import platform

    current_os = platform.system()
    tuple_xy = None
    if current_os == "Windows":
        tuple_xy = _getTerminalSize_windows()
        if tuple_xy is None:
            tuple_xy = _getTerminalSize_tput()
    # needed for window's python in cygwin's xterm!
    if (
        current_os == "Linux"
        or current_os == "Darwin"
        or current_os.startswith("CYGWIN")
    ):
        tuple_xy = _getTerminalSize_linux()
    if tuple_xy is None:
        print("default")
        tuple_xy = (80, 25)  # default value

    return tuple_xy


def _getTerminalSize_windows():
    res = None
    try:
        from ctypes import windll, create_string_buffer

        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12

        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
    except Exception:
        return None
    if res:
        import struct

        (
            bufx,
            bufy,
            curx,
            cury,
            wattr,
            left,
            top,
            right,
            bottom,
            maxx,
            maxy,
        ) = struct.unpack("hhhhHhhhhhh", csbi.raw)
        sizex = right - left + 1
        sizey = bottom - top + 1
        return sizex, sizey
    else:
        return None


def _getTerminalSize_tput():
    # get terminal width
    # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
    try:
        import subprocess

        proc = subprocess.Popen(
            ["tput", "cols"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        output = proc.communicate(input=None)
        cols = int(output[0])
        proc = subprocess.Popen(
            ["tput", "lines"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        output = proc.communicate(input=None)
        rows = int(output[0])
        return (cols, rows)
    except Exception:
        return None


def _getTerminalSize_linux():
    import fcntl
    import termios
    import struct
    import os

    def ioctl_GWINSZ(fd):
        try:
            cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
        except Exception:
            return None
        return cr

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except Exception:
            pass
    if not cr:
        try:
            cr = (os.environ.get("LINES"), os.environ.get("COLUMNS"))
        except Exception:
            return None
    if not cr[0]:
        return None
    return int(cr[1]), int(cr[0])
